# -*- test-case-name: tubes.test.test_protocol -*-
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Objects to connect L{real data <_Protocol>} to L{tubes}.

@see: L{flowFountFromEndpoint}
"""

__all__ = [
    'flowFountFromEndpoint',
    'flowFromEndpoint',
]

from zope.interface import implementer, implementedBy

from .kit import Pauser, beginFlowingFrom, beginFlowingTo, OncePause
from .itube import StopFlowCalled, IDrain, IFount, ISegment
from .listening import Flow

from twisted.python.failure import Failure
from twisted.internet.interfaces import (
    IPushProducer, IListeningPort, IHalfCloseableProtocol
)
from twisted.internet.protocol import Protocol as _Protocol
from twisted.internet.error import ConnectionDone

if 0:
    # Workaround for inability of pydoctor to resolve references.
    from twisted.internet.interfaces import (
        IProtocol, ITransport, IConsumer, IProtocolFactory, IProducer,
        IStreamServerEndpoint
    )
    (IProtocol, ITransport, IConsumer, IProtocolFactory, IProducer,
     IStreamServerEndpoint)
    from twisted.internet.defer import Deferred
    Deferred



@implementer(IPushProducer)
class _FountProducer(object):
    """
    A L{_FountProducer} is an adapter to L{IPushProducer} for an L{IFount}.

    @ivar _fount: An L{IFount}.
    @type _fount: L{IFount}.

    @ivar _pause: A pause if the fount has been paused by C{pauseProducing}
    @type _pause: L{IPause} or L{types.NoneType}
    """
    def __init__(self, fount):
        self._fount = fount
        self._pause = None


    def pauseProducing(self):
        """
        The producer has been paused.  Ensure that the fount is paused.
        """
        # TODO: this implementation is (obviously) incorrect; we could lose
        # track of pauses.  Write some tests.
        self._pause = self._fount.pauseFlow()


    def resumeProducing(self):
        """
        The producer has been resumed.  Ensure that the fount is unpaused.
        """
        self._pause.unpause()


    def stopProducing(self):
        """
        Stop producing data.
        """
        self._fount.stopFlow()



@implementer(IDrain)
class _TransportDrain(object):
    """
    A L{_TransportDrain} is an L{IDrain} that wraps around an object that
    provides L{ITransport} and L{IConsumer}, and delivers data to that
    transport, and flow-control notifications from the consumer.

    @ivar _transport: The transport.
    @type _transport: L{IConsumer} / L{ITransport} provider.
    """

    fount = None
    inputType = ISegment

    def __init__(self, transport):
        self._transport = transport


    def flowingFrom(self, fount):
        """
        Data is flowing to this transport from the given fount.  Register that
        fount as the transport's producer.

        @param fount: the fount producing data - L{ISegment}s - for this
            transport.
        """
        if self.fount is not None:
            self._transport.unregisterProducer()
        beginFlowingFrom(self, fount)
        self._transport.registerProducer(_FountProducer(fount), True)


    def receive(self, item):
        """
        Receive an item of data, some bytes, from the fount.  Pass it along to
        the transport.

        @param item: a fragment of a stream of bytes.
        @type item: L{bytes}
        """
        self._transport.write(item)


    def flowStopped(self, reason):
        """
        The flow of data that should be written to the underlying transport has
        ceased.  Perform a half-close on the transport if possible so that it
        knows no further data is forthcoming.

        @param reason: the reason that the flow stopped; ignored.
        """
        # TODO: this should be loseWriteConnection.
        self._transport.loseConnection()



@implementer(IFount)
class _TransportFount(object):
    """
    An L{IFount} that wraps around an L{ITransport}, and, with the help of a
    L{_ProtocolPlumbing}, delivers any data received by that L{ITransport} to
    an L{IDrain}.

    @ivar _transport: the transport.
    @type _transport: provider of L{ITransport} and L{IProducer}.

    @ivar _pauser: a pauser that will pause and resume the transport.
    @type _pauser: L{Pauser}

    @ivar _preReceivePause: If data is received from the protocol when no drain
        is connected, this will be an L{IPause}.
    @type _preReceivePause: L{IPause} or L{types.NoneType}

    @ivar _preReceiveBuffer: If data is received from the protocol when no
        drain is connected, then this will be the bytes.
    @type _preReceiveBuffer: L{bytes} or L{types.NoneType}
    """

    drain = None
    outputType = ISegment

    def __repr__(self):
        return "<TransportFount for {} paused={}>".format(self._transport,
                                                          self._pauser)


    def __init__(self, transport):
        self._transport = transport
        self._pauser = Pauser(self._transport.pauseProducing,
                              self._transport.resumeProducing)
        self._preReceivePause = None
        self._preReceiveBuffer = None


    def flowTo(self, drain):
        """
        Start delivering data from the transport to the given drain.

        @param drain: the drain that will receive data from the wrapped
            transport.

        @return: the next fount in the chain.
        """
        result = beginFlowingTo(self, drain)
        if self._preReceivePause is not None:
            self._preReceivePause.unpause()
            self.drain.receive(self._preReceiveBuffer)
            self._preReceiveBuffer = None
            self._preReceivePause = None
        return result


    def pauseFlow(self):
        """
        Pause flowing.

        @return: a L{pause token <IPause>}.
        """
        return self._pauser.pause()


    def stopFlow(self):
        """
        End the flow from this fount, dropping the TCP connection in the
        process.
        """
        # Really, stopFlow just ends the *read* connection, but there is no
        # such thing as "loseReadConnection" because TCP can't signal that.
        # This is of potential (academic?) future interest when considering
        # enhanced properties of subprocess transports, because you can both
        # trigger and detect the fact that a subprocess's stdin was closed.
        self._transport.loseConnection()



@implementer(IHalfCloseableProtocol)
class _ProtocolPlumbing(_Protocol):
    """
    An adapter between an L{ITransport} and L{IFount} / L{IDrain} interfaces.

    A L{_ProtocolPlumbing} implements L{IProtocol} to deliver all incoming data
    to the drain associated with its L{fount <IFount>}.

    @ivar _flow: A flow function, as described in L{_factoryFromFlow}.
    @type _flow: L{callable}

    @ivar _drain: The drain that is passed on to the application, created after
        the connection is established in L{_ProtocolPlumbing.connectionMade}.
    @type _drain: L{_TransportDrain}

    @ivar _fount: The fount that is passed on to the application, created after
        the connection is established in L{_ProtocolPlumbing.connectionMade}.
    @type _fount: L{_TransportFount}
    """

    def __init__(self, flow):
        self._flow = flow


    def connectionMade(self):
        """
        The connection was established.  Create an L{IDrain} and an L{IFount}
        and give them to the flow function.
        """
        self._drain = _TransportDrain(self.transport)
        self._fount = _TransportFount(self.transport)
        self._flow(self._fount, self._drain)


    def dataReceived(self, data):
        """
        Some data was received.  Deliver it to the fount created in
        L{connectionMade}.

        @param data: The bytes that were received.
        @type data: L{bytes}
        """
        drain = self._fount.drain
        if drain is None:
            self._fount._preReceivePause = self._fount._pauser.pause()
            self._fount._preReceiveBuffer = data
            return
        drain.receive(data)


    def connectionLost(self, reason):
        """
        The connection was lost.

        If our fount is flowing to a drain, alert that drain that the flow was
        stopped.

        If our drain is flowing from a fount, alert that fount that it should
        stop flowing.

        @param reason: The reason that the connection was terminated.
        @type reason: L{Failure}
        """
        if self._fount.drain is not None:
            self._fount.drain.flowStopped(reason)
        if self._drain.fount is not None:
            self._drain.fount.stopFlow()


    def readConnectionLost(self):
        """
        An end-of-file was received.
        """
        self._fount.drain.flowStopped(Failure(ConnectionDone()))
        self._fount.drain = None


    def writeConnectionLost(self):
        """
        The write output was closed.
        """
        self._drain.fount.stopFlow()
        self._drain.fount = None



def _factoryFromFlow(flow):
    """
    Convert a flow function into an L{IProtocolFactory}.

    A "flow function" is a function which takes a L{fount <IFount>} and an
    L{drain <IDrain>}.

    L{_factoryFromFlow} takes such a function and creates an
    L{IProtocolFactory} which, upon each new connection, provides the flow
    function with an L{IFount} and an L{IDrain} representing the read end and
    the write end of the incoming connection, respectively.

    @param flow: a 2-argument callable, taking (fount, drain).
    @type flow: L{callable}

    @return: a protocol factory.
    @rtype: L{IProtocolFactory}
    """
    from twisted.internet.protocol import Factory
    return Factory.forProtocol(lambda: _ProtocolPlumbing(flow))



@implementer(IFount)
class _FountImpl(object):
    """
    Implementation of fount for listening port.
    """

    outputType = implementedBy(Flow)

    def __init__(self, portObject, aFlowFunction, preListen):
        """
        Create a fount implementation from a provider of L{IPushProducer} and a
        function that takes a fount and a drain.

        @param portObject: the result of the L{Deferred} from
            L{IStreamServerEndpoint.listen}
        @type portObject: L{IListeningPort} and L{IPushProducer} provider
            (probably; workarounds are in place for other cases)

        @param aFlowFunction: a 2-argument callable, invoked when a connection
            arrives, with a fount and drain.
        @type aFlowFunction: L{callable}

        @param preListen: the founts and drains accepted before the C{listen}
            L{Deferred} has fired.  Because these might be arriving before this
            L{_FountImpl} even I{exists}, this needs to be passed in.  That is
            OK because L{_FountImpl} is very tightly coupled to
            L{flowFountFromEndpoint}, which is the only thing that constructs
            it.
        @type preListen: L{list} of 2-L{tuple}s of C{(fount, drain)}
        """
        self.drain = None
        self._preListen = preListen
        self._pauser = Pauser(portObject.pauseProducing,
                              portObject.resumeProducing)
        self._noDrainPause = OncePause(self._pauser)
        self._aFlowFunction = aFlowFunction
        self._portObject = portObject
        if preListen:
            self._noDrainPause.pauseOnce()


    def flowTo(self, drain):
        """
        Start flowing to the given drain.

        @param drain: The drain to send flows to.

        @return: the next fount in the chain.
        """
        result = beginFlowingTo(self, drain)
        self._noDrainPause.maybeUnpause()
        for f, d in self._preListen:
            self._aFlowFunction(f, d)
        return result


    def pauseFlow(self):
        """
        Allow backpressure to build up in the listening socket; ask Twisted to
        stop calling C{accept}.

        @return: An L{IPause}.
        """
        return self._pauser.pause()


    def stopFlow(self):
        """
        Stop the delivery of L{Flow} objects to this L{_FountImpl}'s drain, and
        stop listening on the port represented by this fount.
        """
        self.drain.flowStopped(Failure(StopFlowCalled()))
        self.drain = None
        if IListeningPort.providedBy(self._portObject):
            self._portObject.stopListening()



def flowFountFromEndpoint(endpoint):
    """
    Listen on the given endpoint, and thereby create a L{fount <IFount>} which
    outputs a new L{Flow} for each connection.

    @note: L{IStreamServerEndpoint} formally specifies that its C{connect}
        method returns a L{Deferred} that fires with an L{IListeningPort}.
        However, L{IListeningPort} is insufficient to express the requisite
        flow-control to implement a fount; so the C{endpoint} parameter must be
        an extended endpoint whose C{listen} L{Deferred} fires with a provider
        of both L{IListeningPort} and L{IPushProducer}.  Luckily, the
        real-world implementations of L{IListeningPort} within Twisted are all
        L{IPushProducer}s as well, so practically speaking you will not notice
        this, but for testing it is important to know this is necessary.

    @param endpoint: a server endpoint.
    @type endpoint: L{IStreamServerEndpoint}

    @return: a L{twisted.internet.defer.Deferred} that fires with a L{IFount}
        whose C{outputType} is L{Flow}.
    """
    preListen = []
    def listening(portObject):
        listening.impl = _FountImpl(portObject, aFlowFunction, preListen)
        return listening.impl
    listening.impl = None
    def aFlowFunction(fount, drain):
        if listening.impl is None or listening.impl.drain is None:
            preListen.append((fount, drain))
            if listening.impl is not None:
                listening.impl._noDrainPause.pauseOnce()
        else:
            listening.impl.drain.receive(Flow(fount, drain))
    aFactory = _factoryFromFlow(aFlowFunction)
    return endpoint.listen(aFactory).addCallback(listening)



def flowFromEndpoint(endpoint):
    """
    Convert a client endpoint into a L{Deferred} that fires with a L{Flow}.

    @param endpoint: a client endpoint that will be connected to, once.

    @return: a L{Deferred} that fires with a L{Flow}.
    """
    def cb(fount, drain):
        cb.result = Flow(fount, drain)
    return (endpoint.connect(_factoryFromFlow(cb))
            .addCallback(lambda whatever: cb.result))



# -*- test-case-name: tubes.test.test_tube -*-
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
See L{tube}.
"""

from __future__ import print_function

from zope.interface import implementer
from zope.interface.verify import verifyClass

from twisted.python.components import proxyForInterface
from twisted.python.failure import Failure

from .itube import IDrain, ITube, IDivertable, IFount, StopFlowCalled
from ._siphon import _tubeRegistry, _Siphon, skip
from ._components import _registryActive
from .kit import NoPause as _PlaceholderPause

__all__ = [
    "Diverter",
    "receiver",
    "tube",
    "skip"
]



def tube(cls):
    """
    L{tube} is a class decorator which declares a given class to be an
    implementer of L{ITube} and fills out any methods or attributes which are
    not present on the decorated type with null-implementation methods (those
    which return None) and None attributes.

    @param cls: A class with some or all of the attributes or methods described
        by L{ITube}.
    @type cls: L{type}

    @return: C{cls}
    @rtype: L{type} which implements L{ITube}
    """

    # This is better than a superclass, because:

    # - you can't do a separate 'isinstance(Tube)' check instead of
    #   ITube.providedBy like you're supposed to

    # - you can't just instantiate Tube directly, that is pointless
    #   functionality so we're not providing it

    # - it avoids propagating a bad example that other codebases will copy to
    #   depth:infinity, rather than depth:1 where subclassing is actually sort
    #   of okay

    # - it provides a more straightforward and reliable mechanism for
    #   future-proofing code.  If you're inheriting from a superclass and you
    #   want it to do something to warn users, upgrade an interface, and so on,
    #   you have to try to cram a new meta-type into the user's hierarchy so a
    #   function gets invoked at the right time.  If you're invoking this class
    #   decorator, then it just gets invoked like a normal function, and we can
    #   put some code in here that examines the type and does whatever it wants
    #   to do, because the @ syntax simply called it as a function.

    # It still shares some issues with inheritance, such as:

    # - the direction of visibility within the hierarchy is still wrong.  you
    #   can still do 'self.someMethodIDidntImplement()' and get a result.

    # - it destructively modifies the original class, so what you see isn't
    #   quite what you get.  a cleaner compositional approach would simply wrap
    #   an object around another object (but that would mean inventing a new
    #   incompletely-specified type that floats around at runtime, rather than
    #   a utility to help you completely implement ITube at import time)

    def started(self):
        """
        A null implementation of started.

        @param self: An instance of the C{tube} being defined.
        """

    def stopped(self, reason):
        """
        A null implementation of stopped.

        @param self: An instance of the C{tube} being defined.

        @param reason: see L{ITube}
        """

    def received(self, item):
        """
        A null implementation of received.

        @param self: An instance of the C{tube} being defined.

        @param item: see L{ITube}
        """

    fillers = [('started', started),
               ('stopped', stopped),
               ('received', received),
               ('inputType', None),
               ('outputType', None)]

    notHere = object()

    for name, value in fillers:
        if getattr(cls, name, notHere) is notHere:
            setattr(cls, name, value)

    cls = implementer(ITube)(cls)
    verifyClass(ITube, cls)
    return cls



@implementer(ITube)
class _Tubule(object):
    """
    A tube created for the C{@tube} decorator.
    """
    def __init__(self, inputType, outputType, received, name):
        """
        @param inputType: An interface for the input type.

        @param outputType: an interface for the output type.

        @param received: a callable to implement C{received}.

        @param name: a string describing this L{_Tubule}.
        """
        self.inputType = inputType
        self.outputType = outputType
        self.received = received
        self._name = name


    def started(self):
        """
        Tubules cannot produce a greeting.

        @return: an empty iterable.
        """
        return ()


    def stopped(self, reason):
        """
        Tubules cannot produce a farewell.

        @param reason: the reason the flow stopped.

        @return: an empty iterable.
        """
        return ()


    def __repr__(self):
        """
        @return: this L{_Tubule}'s name.
        """
        return self._name



def receiver(inputType=None, outputType=None, name=None):
    """
    Decorator for a stateless function which receives inputs.

    For example, to add 1 to each in a stream of numbers::

        @receiver(inputType=int, outputType=int)
        def addOne(item):
            yield item + 1

    @param inputType: The C{inputType} attribute of the resulting L{ITube}.

    @param outputType: The C{outputType} attribute of the resulting L{ITube}.

    @param name: a name describing the tubule for it to show as in a C{repr}.
    @type name: native L{str}

    @return: a stateless tube with the decorated method as its C{received}
        method.
    @rtype: L{ITube}
    """
    def decorator(decoratee):
        return _Tubule(inputType, outputType, decoratee,
                       name if name is not None else decoratee.__name__)
    return decorator



def series(start, *tubes):
    """
    Connect up a series of objects capable of transforming inputs to outputs;
    convert a sequence of L{ITube} objects into a sequence of connected
    L{IFount} and L{IDrain} objects.  This is necessary to be able to C{flowTo}
    an object implementing L{ITube}.

    This function can best be understood by understanding that::

        x = a
        a.flowTo(b).flowTo(c)

    is roughly analagous to::

        x = series(a, b, c)

    with the additional feature that C{series} will convert C{a}, C{b}, and
    C{c} to the requisite L{IDrain} objects first.

    @param start: The initial element in the chain; the object that will
        consume inputs passed to the result of this call to C{series}.
    @type start: an L{ITube}, or anything adaptable to L{IDrain}.

    @param tubes: Each element of C{plumbing}.
    @type tubes: a L{tuple} of L{ITube}s or objects adaptable to L{IDrain}.

    @return: An L{IDrain} that can consume inputs of C{start}'s C{inputType},
        and whose C{flowingFrom} will return an L{IFount} that will produce
        outputs of C{plumbing[-1]} (or C{start}, if plumbing is empty).
    @rtype: L{IDrain}

    @raise TypeError: if C{start}, or any element of C{plumbing} is not
        adaptable to L{IDrain}.
    """
    with _registryActive(_tubeRegistry):
        result = IDrain(start)
        currentFount = result.flowingFrom(None)
        drains = [IDrain(tube) for tube in tubes]
    for drain in drains:
        currentFount = currentFount.flowTo(drain)
    return result



@tube
class _DrainingTube(object):
    """
    A L{_DrainingTube} is an L{ITube} that unbuffers a list of items.  It is an
    implementation detail of the way that L{Diverter} works.
    """

    def __init__(self, items, eventualUpstream, eventualDownstream):
        """
        Create a L{_DrainingTube} with some C{items} to drain, a L{drain
        <IDrain>} to drain them to, and a L{fount <IFount>} to flow to that
        C{drain} once the items are flowed.

        @param items: An iterable of items to drain.
        @type items: iterable

        @param eventualUpstream: a L{fount <IFount>} which should flow to
            C{eventualDownstream} once the last item in C{items} has been
            passed on.

        @param eventualDownstream: a L{drain <IDrain>} which should receive
            each item in C{items} and then accept the flow from
            C{eventualUpstream}.
        """
        self._items = list(items)
        self._eventualUpstream = eventualUpstream
        self._eventualDownstream = eventualDownstream


    def __repr__(self):
        """
        Display the remaining items to be drained.
        """
        return ("<Draining Tube {0}>".format(repr(self._items)))


    def started(self):
        """
        Yield each item from the C{items} passed to the constructor, then
        switch flow to C{_eventualUpstream}.
        """
        while self._items:
            item = self._items.pop(0)
            yield item
        self._eventualUpstream.flowTo(self._eventualDownstream)



@implementer(IFount)
class _NullFount(object):
    """
    An I{almost} no-op implementation of fount which does nothing but update
    its C{drain} to point at itself.
    """

    outputType = None
    drain = None

    def flowTo(self, drain):
        """
        Update the C{drain} attribute of this L{_NullFount} and call
        C{flowingFrom} on the given C{drain}.

        @param drain: see L{IFount}

        @return: see L{IFount}
        """
        self.drain = drain
        return drain.flowingFrom(self)


    def pauseFlow(self):
        """
        Return an L{IPause} which does nothing, and then does nothing when
        resumed.

        @return: see L{IFount}
        """
        return _PlaceholderPause()


    def stopFlow(self):
        """
        Do nothing.
        """
        self.drain.flowStopped(Failure(StopFlowCalled()))



class Diverter(proxyForInterface(IDrain, "_drain")):
    """
    A L{Diverter} is a L{drain <IDrain>} which maintains a buffer of items not
    yet received by its L{IDivertable} down-stream drain.
    """

    def __init__(self, divertable):
        """
        Create a L{Diverter}.

        @param divertable: Divert a divertable.
        @type divertable: L{IDivertable} provider
        """
        if not IDivertable.providedBy(divertable):
            raise TypeError("Diverter can only wrap IDivertable providers.")
        self._divertable = divertable
        self._friendSiphon = _Siphon(divertable)
        self._drain = self._friendSiphon._tdrain


    def __repr__(self):
        """
        Nice string representation for this Diverter which mentions what it is
        diverting.
        """
        return "<Diverter for {0}>".format(self._divertable)


    def divert(self, drain):
        """
        Divert the flow from the fount which is flowing I{into this diverter}
        to instead flow into I{the given drain}, reassembling any buffered
        output from this siphon's tube first.

        @param drain: The L{drain <IDrain>} to divert the flow I{to}.
        @type drain: L{IDrain}

        @return: L{None}
        """
        unpending = self._friendSiphon.ejectPending()

        pendingPending = self._divertable.reassemble(unpending) or []
        upstream = self._friendSiphon._tdrain.fount
        nullFount = _NullFount()
        dt = series(_DrainingTube(pendingPending, upstream, drain))
        again = nullFount.flowTo(dt)
        again.flowTo(drain)



# -*- test-case-name: tubes.test.test_undefer -*-
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Tests for L{tubes.undefer}.
"""

from twisted.trial.unittest import SynchronousTestCase
from twisted.internet.defer import succeed
from twisted.internet.defer import Deferred
from twisted.python.failure import Failure

from tubes.undefer import deferredToResult, fountToDeferred
from tubes.test.util import FakeDrain
from tubes.test.util import FakeFount
from tubes.tube import tube, series

class DeferredIntegrationTests(SynchronousTestCase):
    """
    Tests for L{deferredToResult}.
    """
    def setUp(self):
        """
        Create a fount and drain.
        """
        self.ff = FakeFount()
        self.fd = FakeDrain()


    def test_tubeYieldsFiredDeferred(self):
        """
        When a tube yields a fired L{Deferred} its result is synchronously
        delivered.
        """

        @tube
        class SucceedingTube(object):
            def received(self, data):
                yield succeed(''.join(reversed(data)))

        fakeDrain = self.fd
        self.ff.flowTo(series(SucceedingTube(),
                              deferredToResult())).flowTo(fakeDrain)
        self.ff.drain.receive("hello")
        self.assertEqual(self.fd.received, ["olleh"])


    def test_tubeYieldsUnfiredDeferred(self):
        """
        When a tube yields an unfired L{Deferred} its result is asynchronously
        delivered.
        """

        d = Deferred()

        @tube
        class WaitingTube(object):
            def received(self, data):
                yield d

        fakeDrain = self.fd
        self.ff.flowTo(series(WaitingTube(),
                              deferredToResult())).flowTo(fakeDrain)
        self.ff.drain.receive("ignored")
        self.assertEqual(self.fd.received, [])

        d.callback("hello")

        self.assertEqual(self.fd.received, ["hello"])


    def test_tubeYieldsMultipleDeferreds(self):
        """
        When a tube yields multiple deferreds their results should be delivered
        in order.
        """

        d = Deferred()

        @tube
        class MultiDeferredTube(object):
            didYield = False
            def received(self, data):
                yield d
                MultiDeferredTube.didYield = True
                yield succeed("goodbye")

        fakeDrain = self.fd
        self.ff.flowTo(series(MultiDeferredTube(),
                              deferredToResult())).flowTo(fakeDrain)
        self.ff.drain.receive("ignored")
        self.assertEqual(self.fd.received, [])

        d.callback("hello")

        self.assertEqual(self.fd.received, ["hello", "goodbye"])


    def test_tubeYieldedDeferredFiresWhileFlowIsPaused(self):
        """
        When a L{Tube} yields an L{Deferred} and that L{Deferred} fires when
        the L{_SiphonFount} is paused it should buffer it's result and deliver
        it when L{_SiphonFount.resumeFlow} is called.
        """
        d = Deferred()

        @tube
        class DeferredTube(object):
            def received(self, data):
                yield d

        fakeDrain = self.fd
        self.ff.flowTo(series(DeferredTube(),
                              deferredToResult())).flowTo(fakeDrain)
        self.ff.drain.receive("ignored")

        anPause = self.fd.fount.pauseFlow()

        d.callback("hello")
        self.assertEqual(self.fd.received, [])

        anPause.unpause()
        self.assertEqual(self.fd.received, ["hello"])


    def test_tubeStoppedDeferredly(self):
        """
        The L{_Siphon} stops its L{Tube} and propagates C{flowStopped}
        downstream upon the completion of all L{Deferred}s returned from its
        L{Tube}'s C{stopped} implementation.
        """
        reasons = []
        conclusion = Deferred()
        @tube
        class SlowEnder(object):
            def stopped(self, reason):
                reasons.append(reason)
                yield conclusion

        self.ff.flowTo(series(SlowEnder(), deferredToResult(), self.fd))
        self.assertEqual(reasons, [])
        self.assertEqual(self.fd.received, [])

        stopReason = Failure(ZeroDivisionError())

        self.ff.drain.flowStopped(stopReason)
        self.assertEqual(self.fd.received, [])
        self.assertEqual(len(reasons), 1)
        self.assertIdentical(reasons[0].type, ZeroDivisionError)
        self.assertEqual(self.fd.stopped, [])

        conclusion.callback("conclusion")
        # Now it's really done.
        self.assertEqual(self.fd.received, ["conclusion"])
        self.assertEqual(self.fd.stopped, [stopReason])


    def test_fountToDeferred(self):
        """
        L{fountToDeferred} returns a L{Deferred} that fires with an iterable of
        all the objects that the fount passed to it emits.
        """
        self.assertIsNone(self.ff.drain)
        d = fountToDeferred(self.ff)
        self.assertIsNotNone(self.ff.drain)
        self.assertNoResult(d)
        self.ff.drain.receive(1)
        self.assertNoResult(d)
        self.ff.drain.receive(2)
        self.ff.drain.flowStopped(Failure(ZeroDivisionError()))
        self.assertEqual(list(self.successResultOf(d)), [1, 2])



# -*- test-case-name: tubes.test.test_kit -*-
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Tests for L{tubes.kit}.
"""

from zope.interface.verify import verifyObject

from twisted.trial.unittest import SynchronousTestCase as TestCase

from ..itube import IPause, AlreadyUnpaused
from ..kit import Pauser

def countingCallable():
    """
    Generate a callable for testing.

    @return: a callable with a C{d} attribute that indicates the number of
        times it's been called.
    """
    def callee():
        callee.d += 1
    callee.d = 0
    return callee



class PauserTests(TestCase):
    """
    Tests for L{Pauser}, helper for someone who wants to implement a thing
    that pauses.
    """

    def test_pauseOnce(self):
        """
        One call to L{_Pauser.pause} will call the actuallyPause callable.
        """
        pause = countingCallable()
        pauser = Pauser(pause, None)
        result = pauser.pause()
        self.assertTrue(verifyObject(IPause, result))
        self.assertEqual(pause.d, 1)


    def test_pauseThenUnpause(self):
        """
        A call to L{_Pauser.pause} followed by a call to the result's
        C{unpause} will call the C{actuallyResume} callable.
        """
        pause = countingCallable()
        resume = countingCallable()
        pauser = Pauser(pause, resume)
        pauser.pause().unpause()
        self.assertEqual(pause.d, 1)
        self.assertEqual(resume.d, 1)


    def test_secondUnpauseFails(self):
        """
        The second of two consectuive calls to L{IPause.unpause} results in an
        L{AlreadyUnpaused} exception.
        """
        pause = countingCallable()
        resume = countingCallable()
        pauser = Pauser(pause, resume)
        aPause = pauser.pause()
        aPause.unpause()
        self.assertRaises(AlreadyUnpaused, aPause.unpause)
        self.assertEqual(resume.d, 1)


    def test_repeatedlyPause(self):
        """
        Multiple calls to L{_Pauser.pause} where not all of the pausers are
        unpaused do not result in any calls to C{actuallyResume}.
        """
        pause = countingCallable()
        resume = countingCallable()
        pauser = Pauser(pause, resume)
        one = pauser.pause()
        two = pauser.pause()
        three = pauser.pause()
        four = pauser.pause()

        one.unpause()
        two.unpause()
        three.unpause()
        self.assertEqual(pause.d, 1)
        self.assertEqual(resume.d, 0)
        four.unpause()
        self.assertEqual(resume.d, 1)


    def test_reentrantPause(self):
        """
        A L{Pauser} that pauses re-entrantly will only result in one call to
        the specified C{pause} callable.
        """
        def pause():
            pause.d += 1
            pauser.pause()
        pause.d = 0
        resume = countingCallable()
        pauser = Pauser(pause, resume)
        pauser.pause()
        self.assertEqual(pause.d, 1)
        self.assertEqual(resume.d, 0)


    def test_reentrantResume(self):
        """
        A L{Pauser} that resumes re-entrantly will raise L{AlreadyUnpaused}.
        """
        pause = countingCallable()
        def resume():
            resume.d += 1
            self.assertRaises(AlreadyUnpaused, anPause.unpause)
        resume.d = 0
        pauser = Pauser(pause, resume)
        anPause = pauser.pause()
        anPause.unpause()
        anotherPause = pauser.pause()
        self.assertEqual(resume.d, 1)
        self.assertEqual(pause.d, 2)
        anotherPause.unpause()
        self.assertEqual(resume.d, 2)



# -*- test-case-name: tubes.test.test_listening -*-
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Tests for L{tubes.listening}.
"""

from unittest import TestCase

from ..listening import Flow, Listener
from ..memory import iteratorFount

from .util import FakeDrain

class ListeningTests(TestCase):
    """
    Test cases for listening.
    """

    def test_listenerCallsFlowConnector(self):
        """
        A L{Listener} is a drain which calls the function given to it to
        connect a flow
        """
        drained = FakeDrain()
        flow = Flow(iteratorFount([1, 2, 3]),
                    drained)
        flows = []
        fi = iteratorFount([flow])
        listener = Listener(flows.append)
        fi.flowTo(listener)
        self.assertEqual(len(flows), 1)
        results = FakeDrain()
        flows[0].fount.flowTo(results)
        # The listener might need to (and in fact does) interpose a different
        # value for 'fount' and 'drain' to add hooks to them.  We assert about
        # the values passed through them.
        self.assertEqual(results.received, [1, 2, 3])
        iteratorFount([4, 5, 6]).flowTo(flows[0].drain)
        self.assertEqual(drained.received, [4, 5, 6])


    def test_listenerLimitsConcurrentConnections(self):
        """
        L{Listener} will pause its fount when too many connections are
        received.
        """
        connectorCalled = []
        listener = Listener(connectorCalled.append, maxConnections=3)
        tenFlows = iteratorFount([Flow(iteratorFount([1, 2, 3]),
                                       FakeDrain())
                                  for each in range(10)])
        tenFlows.flowTo(listener)
        self.assertEqual(len(connectorCalled), 3)
        connectorCalled[0].fount.flowTo(connectorCalled[0].drain)
